@@ -91,17 +91,33 @@ static void setTriple(Options& opts)
   } else if (pd.find("#define __i386__ 1") != pd.npos ||
              pd.find("#define _M_IX86 ") != pd.npos) {
     triple.setArchName("i386");
+  } else if (pd.find("#define __aarch64__ 1") != pd.npos ||
+             pd.find("#define _M_ARM64 ") != pd.npos) {
+    triple.setArchName("aarch64");
   }
   if (pd.find("#define _WIN32 1") != pd.npos) {
     triple.setVendorName("pc");
     triple.setOSName("windows");
   }
+  if (pd.find("#define _MSC_VER ") != pd.npos) {
+    triple.setEnvironmentName("msvc");
+  }
   if (pd.find("#define __MINGW32__ 1") != pd.npos) {
     triple.setEnvironmentName("gnu");
   }
   opts.Triple = triple.getTriple();
 }
 
+static bool isBuiltinIncludeDir(std::string const& inc)
+{
+  // FIXME: Intrinsics headers are platform-specific.
+  // Is there a better way to detect this directory?
+  return (llvm::sys::fs::exists(inc + "/emmintrin.h")   // x86_64
+          || llvm::sys::fs::exists(inc + "/altivec.h")  // ppc64
+          || llvm::sys::fs::exists(inc + "/arm_neon.h") // aarch64
+  );
+}
+
 static bool detectCC_GNU(const char* const* argBeg, const char* const* argEnd,
                          Options& opts, const char* id, const char* ext)
 {
@@ -143,11 +159,7 @@ static bool detectCC_GNU(const char* const* argBeg, const char* const* argEnd,
                      fwImplicitSuffix));
             }
             // Replace the compiler builtin include directory with ours.
-            if (!fw &&
-                // FIXME: Intrinsics headers are platform-specific.
-                // Is there a better way to detect this directory?
-                (llvm::sys::fs::exists(inc + "/emmintrin.h") ||
-                 llvm::sys::fs::exists(inc + "/altivec.h"))) {
+            if (!fw && isBuiltinIncludeDir(inc)) {
               inc = getClangBuiltinIncludeDir();
             }
             opts.Includes.push_back(Options::Include(inc, fw));
