@@ -5,7 +5,7 @@
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
 
-      http://www.apache.org/licenses/LICENSE-2.0
+      https://www.apache.org/licenses/LICENSE-2.0
 
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
@@ -56,6 +56,12 @@
 #  define CASTXML_OWNS_OSTREAM
 #endif
 
+#if LLVM_VERSION_MAJOR > 9
+#  define CASTXML_MAKE_UNIQUE std::make_unique
+#else
+#  define CASTXML_MAKE_UNIQUE llvm::make_unique
+#endif
+
 class ASTConsumer : public clang::ASTConsumer
 {
   clang::CompilerInstance& CI;
@@ -281,6 +287,17 @@ protected:
                     "#define __float128 __castxml__float128\n";
       }
 
+      if (CI.getLangOpts().MSCompatibilityVersion >= 192300000) {
+        // MSVC tools 14.23 and above declare __builtin_assume_aligned
+        // in "intrin0.h" with a signature incompatible with Clang's
+        // builtin.  It is guarded by '!defined(__clang__)' but that
+        // breaks when we simulate MSVC's preprocessor.  Rename the
+        // builtin via preprocessing to avoid the conflict.
+        builtins += "\n"
+                    "#define __builtin_assume_aligned"
+                    " __castxml__builtin_assume_aligned\n";
+      }
+
       // Provide __is_assignable builtin if simulating MSVC.
       // When a future Clang version supports the builtin then
       // we can skip this when built against such a Clang.
@@ -351,6 +368,14 @@ protected:
                     "#define __NO_MATH_INLINES 1\n";
       }
 
+      // Tolerate glibc use of a GNU extension not implemented by Clang.
+      if (this->NeedAttributeMallocArgs(this->Opts.Predefines)) {
+        // Clang does not support '__attribute__((__malloc__(args...)))'
+        // used in glibc 2.34+ headers.
+        builtins += "\n"
+                    "#define __malloc__(...) __malloc__\n";
+      }
+
     } else {
       builtins += predefines.substr(start, end - start);
     }
@@ -371,6 +396,11 @@ protected:
     return this->IsActualGNU(pd);
   }
 
+  bool NeedAttributeMallocArgs(std::string const& pd)
+  {
+    return this->IsActualGNU(pd);
+  }
+
   bool NeedFloat128(std::string const& pd) const
   {
     return (this->IsActualGNU(pd) &&
@@ -443,11 +473,11 @@ class CastXMLSyntaxOnlyAction
 #ifdef CASTXML_OWNS_OSTREAM
     } else if (std::unique_ptr<llvm::raw_ostream> OS =
                  CI.createDefaultOutputFile(false, filename(InFile), "xml")) {
-      return llvm::make_unique<ASTConsumer>(CI, std::move(OS), this->Opts);
+      return CASTXML_MAKE_UNIQUE<ASTConsumer>(CI, std::move(OS), this->Opts);
 #else
     } else if (llvm::raw_ostream* OS =
                  CI.createDefaultOutputFile(false, filename(InFile), "xml")) {
-      return llvm::make_unique<ASTConsumer>(CI, *OS, this->Opts);
+      return CASTXML_MAKE_UNIQUE<ASTConsumer>(CI, *OS, this->Opts);
 #endif
     } else {
       return nullptr;
@@ -534,8 +564,12 @@ runClangCreateDiagnostics(const char* const* argBeg, const char* const* argEnd)
     new clang::DiagnosticOptions);
   llvm::IntrusiveRefCntPtr<clang::DiagnosticIDs> diagID(
     new clang::DiagnosticIDs());
+#if LLVM_VERSION_MAJOR >= 10
+  llvm::opt::OptTable const* opts = &clang::driver::getDriverOptTable();
+#else
   std::unique_ptr<llvm::opt::OptTable> opts(
     clang::driver::createDriverOptTable());
+#endif
   unsigned missingArgIndex, missingArgCount;
 #if LLVM_VERSION_MAJOR > 3 ||                                                 \
   LLVM_VERSION_MAJOR == 3 && LLVM_VERSION_MINOR >= 7
@@ -612,7 +646,13 @@ static int runClangImpl(const char* const* argBeg, const char* const* argEnd,
       const char* const* cmdArgBeg = cmd->getArguments().data();
       const char* const* cmdArgEnd = cmdArgBeg + cmd->getArguments().size();
       if (clang::CompilerInvocation::CreateFromArgs(
-            CI->getInvocation(), cmdArgBeg, cmdArgEnd, *diags)) {
+            CI->getInvocation(),
+#if LLVM_VERSION_MAJOR > 9
+            llvm::makeArrayRef(cmdArgBeg, cmdArgEnd),
+#else
+            cmdArgBeg, cmdArgEnd,
+#endif
+            *diags)) {
         if (diags->hasErrorOccurred()) {
           return 1;
         }
